"""
Date

This module contains actions used to get dates & times.
"""

import datetime
import functools
from typing import TypedDict, Union, Callable


class Date(TypedDict):
    r"""
- Contains date & time information

- Used as an _output_ from [now](../ppa_tools.date/#now)

!!! example

    ```yaml linenums="1"
    display: Tue Aug 25 08:10:45 2020
    date: 25/08/2020
    time: '08:10:45'
    day: 25
    day_name: Tuesday
    month: 8
    month_name: August
    year: 2020
    hour: 8
    minutes: 10
    seconds: 45
    ```

!!! info "Date Keys"

    __display__: The day, month, date, time, & year displayed together

    __date__: The date in __dd/mm/yyyy__ format

    __time__: The time in __hh:mm:ss__ format

    __day__: The day as a __number__

    __day_name__: The name of the day

    __month__: The month as a __number__

    __month_name__: The name of the month

    __year__: The year as a __number__

    __hour__: The hour as a __number__

    __minutes__: The minutes as a __number__

    __seconds__: The seconds as a __number__

    __timestamp__: The date & time as a timestamp

    """
    display: str
    date: str
    time: str
    day: int
    day_name: str
    month: int
    month_name: str
    year: int
    hour: int
    minutes: int
    seconds: int
    timestamp: int


def _generate_date_dict(func: Callable) -> Date:
    @functools.wraps(func)
    def _fn(*args, **kwargs):
        date_object = func(*args, **kwargs)
        date = {
            "display": date_object.strftime("%c"),
            "date": date_object.strftime("%d/%m/%Y"),
            "time": date_object.strftime("%H:%M:%S"),
            "day": int(date_object.strftime("%d")),
            "day_name": date_object.strftime("%A"),
            "month": int(date_object.strftime("%m")),
            "month_name": date_object.strftime("%B"),
            "year": int(date_object.strftime("%Y")),
            "hour": int(date_object.strftime("%H")),
            "minutes": int(date_object.strftime("%M")),
            "seconds": int(date_object.strftime("%S")),
            "timestamp": int(datetime.datetime.timestamp(date_object)),
        }
        return Date(**date)

    return _fn


@_generate_date_dict
def now() -> Date:
    """Get the current ___UTC___ date & time.

    !!! question "Minimum Plugin Version: 3.0.0"

    ??? info "Input"

        This action takes no inputs.

    ??? success "Output"

        A single [Date](../inputs_outputs/#date) dictionary.

    ??? example

        ```YAML linenums="1"
        - ppa_tools.date.now:
          save: date
        ```
    """
    return datetime.datetime.utcnow()


@_generate_date_dict
def from_timestamp(timestamp: Union[str, int]) -> Date:
    """Convert a UNIX Epoch timestamp into a Date dictionary

    !!! question "Minimum Plugin Version: 3.2.0"

    ??? info "Input"

        __timestamp:__ The UNIX Epoch timestamp to convert

    ??? success "Output"

        A single [Date](../inputs_outputs/#date) dictionary.

    ??? example

        ```YAML linenums="1"
        - ppa_tools.date.from_timestamp:
            timestamp: 1601556657
          save: date
        ```
    """
    return datetime.datetime.fromtimestamp(int(timestamp))


@_generate_date_dict
def add_seconds_to_timestamp(timestamp: Union[str, int], seconds: Union[str, int]) -> Date:
    """Add a number of seconds to a UNIX Epoch timestamp value

    !!! question "Minimum Plugin Version: 3.2.0"

    ??? info "Input"

        __timestamp:__ The UNIX Epoch timestamp to add the seconds to

        __seconds:__ The seconds to add to the timestamp


    ??? success "Output"

        A updated [Date](../inputs_outputs/#date) dictionary with the seconds added to it.

    ??? example

        ```YAML linenums="1"
        - ppa_tools.date.add_seconds_to_timestamp:
            timestamp: 1601556657
            seconds: 500
          save: date
        ```
    """
    return datetime.datetime.fromtimestamp(int(timestamp)) + datetime.timedelta(seconds=int(seconds))


@_generate_date_dict
def subtract_seconds_from_timestamp(timestamp: Union[str, int], seconds: Union[str, int]) -> Date:
    """Subtract a number of seconds from a UNIX Epoch timestamp value

    !!! question "Minimum Plugin Version: 3.2.0"

    ??? info "Input"

        __timestamp:__ The UNIX Epoch timestamp to remove the seconds from

        __seconds:__ The seconds to subtract from the timestamp


    ??? success "Output"

        A updated [Date](../inputs_outputs/#date) dictionary with the seconds subtracted from it.

    ??? example

        ```YAML linenums="1"
        - ppa_tools.date.subtract_seconds_from_timestamp:
            timestamp: 1601556657
            seconds: 500
          save: date
        ```
    """
    return datetime.datetime.fromtimestamp(int(timestamp)) - datetime.timedelta(seconds=int(seconds))


def days_between(start_timestamp: Union[str, int], end_timestamp: Union[str, int]) -> int:
    """Get the number of days between two timestamps

    !!! question "Minimum Plugin Version: 3.2.0"

    ??? info "Input"

        __start_timestamp:__ The earliest of the two UNIX Epoch timestamps

        __end_timestamp:__ The latest of the two UNIX Epoch timestamps


    ??? success "Output"

        An integer.

    ??? example

        ```YAML linenums="1"
        - ppa_tools.date.days_between:
            start_timestamp: 1601556657
            end_timestamp: 1601653847
          save: date
        ```
    """
    start_date = datetime.datetime.utcfromtimestamp(int(start_timestamp))
    end_date = datetime.datetime.utcfromtimestamp(int(end_timestamp))
    return int((end_date - start_date).days)
